#! /usr/bin/env python
#-----------------------------------------------------------------------------
# Name:        midi2c.py
# Purpose:     A program convert midi (.mid) files to
#              C include data files.
#
# Author:      Tom van Zutphen, tom(@)thingz(.)nl.
#
# Created:     2011/04/14
# RCS-ID:      $Id: midi2c.py $
# Copyright:   (c) 2011
# Licence:     GPL
#-----------------------------------------------------------------------------
__revision__ = '$Id: midi2c.py,v 0.70 2011/05/04 10:05 Tom van Zutphen Exp $'
"""
Midi file converter

A script to convert midi (.mid) files to C code include Data files.
Extracts time and the notes to play
(C)2011 Tom van Zutphen, tom(@)thingz(.)nl.

Known bugs (ToDO): for not note events deltatime is added until next note. Afterwards this combined deltatime is converted to (mili)seconds.
There could be an error here since part of the combined delta times could have another tempo setting which would then be a different time in (mili)seconds.
This could be noticable if the seperate tracks/channels with (mili)second timeing are combined again in playing (they will not be synchronised anymore).
If not played together the possible time difference is not noticable i guess.

"""

#import string
import anydbm,os.path,re
import math
#import binascii
#import array
#import tempfile

def cls():
    a=os.name
    if a=='posix': #linux
        os.system('clear')
    elif a=='nt': #windows
        os.system('cls')
    else: #unknown
        print

class Midi_Base:
    " Docstring............."
    def __init__(self, infilename,bin):
        self.bin=bin
        self.instruments=['','','','','','','','','','','','','','','','']
        self.numerator=4   #not realy needed for time calc
        self.denominator=4 #do we need his for time calc?
        self.denominatorHeader='\nint denominator = '
        self.denominatorFooter=';'
        self.bar='\n// Time signature(s): '        #just there to keep track of al time signature events (the above only remembers the last one or default one)
        self.ppqn=0 #Delta-time ticks per quarter note
        self.noTracks=0
        self.trackdata={}
        self.bpmtable=[]
        try:
            mididata = open(infilename)
            header = mididata.read(8)
            if header == "\x4D\x54\x68\x64\x00\x00\x00\x06":                #Default midi header
                print "Default midi header detected, seems we're good to go..."
                header = mididata.read(1)
                header = ord(mididata.read(1))
                if (header == 0): print "0 - Single track midifile"
                elif (header == 1): print "1 - multiple tracks, synchronous midifile"
                elif (header == 2): print "2 - multiple tracks, asynchronous midifile"
                else: print header," - Unknown type midifile. this may couse some trouble."
                self.noTracks=ord(mididata.read(1))*256+ord(mididata.read(1))
                print "Number of tracks: ",self.noTracks
                header = ord(mididata.read(1))
                if header>=128:              #first byte is negative, (positive) represents frames per second (-25, -25, -29, -30 according to SMPTE)
                   header=256-header
                   self.ppqn=header*ord(mididata.read(1)) #second byte is resolution of frame, together the are the total ticks per second example: first byte: -25 = 25 frames per second, 2nd byte 40, 40 tickt per frame. total 25*40 is 1000 ticks per second is 1ms per tick
                   print "Number of delta-time ticks per second:",self.ppqn
                else:
                   self.ppqn=header*256+ord(mididata.read(1)) #First byte = positive (or 0), the 2 bytes represent the number of ticks per quarter note (default 120bpm = 0,5 sec per quarter note
                print "Number of delta-time ticks per quarter note:",self.ppqn
            else:
                cls()
                print 'Unknown file format, this is not a midi file i can handle....\n'
                sys.exit()
        except IOError:

            print 'file ' + infilename + ' cannot be opened!\n'
            sys.exit()
    def closeTrack(self,track,channel,lostuS,compensateduS,data,datalen,outfilename,bin):
        if (lostuS):
           print "Total lost time in uS for last track: ",lostuS," Compensated time in uS: ",compensateduS
        if (bin):
           import struct
           writer = file(outfilename+"T"+str(track)+"C"+str(channel)+'.mmi', 'wb')
           for byte in data:
              writer.write(struct.pack('B', byte))
        else:
           writer = file(outfilename+"_T"+str(track)+"C"+str(channel)+'.h', 'w')
           writer.write('#ifndef _inc_midifile_\n#define _inc_midifile_\n\n')
           writer.write('// Songname with Track and Channel nrs: '+outfilename+"_T"+str(track)+"C"+str(channel)+'\n')
           writer.write('//#include "songs/'+outfilename+'_T'+str(track)+'C'+str(channel)+'.h"\n')
           writer.write('PROGMEM prog_uint8_t midisong[')
           writer.write(str(datalen[channel]))
           writer.write(']={\n')
           writer.write(data)
           writer.write('  };\n')
           writer.write(self.bar)
           writer.write(self.denominatorHeader+str(self.denominator)+self.denominatorFooter)
           writer.write('\n// used instruments:')
           writer.write(self.instruments[channel])
           writer.write('\n#endif /*_midifile_*/\n')
        writer.close()
        print "\n\nfile: "+outfilename+"_T"+str(track)+"C"+str(channel)
#        writer.write('// Songname with Track and Channel nrs: '+outfilename+"_T"+str(track)+"C"+str(channel)+'\n')
#        print "Length: "+str(datalen[channel])
        print self.bar+" ppqn from header: ",self.ppqn
#        print self.denominatorHeader+str(self.denominator)+self.denominatorFooter
        print "used instruments: "+self.instruments[channel]+"\n\n"


    def deltaTics2deltaMiliSecs(self,maxlen,outfilename):
      for track in self.trackdata:
         for channel in self.trackdata[track]:
            data=self.trackdata[track][channel]
            start=[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]    # keep track of nr of data bytes on 1 line of c code
            datalen=[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]
            totaltimeOld=[0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0]
            mpqnT=500000
            new_mpqnT=0
            totaltimeInMicrosecondsT=0.0
            totaltimeInMicrosecondsTOld=0.0
            deltaTimeNote_microsecsT=0.0
            bpmindexT=0
            lostuS=0.0
            compensateduS=0.0
            totaltime=0
            flap=0
            flapold=0
            mindelta=0
            maxdelta=0
            if (bin):
               dataout=[]
            else:
               dataout=""
            for note in data:
               delta_event=note[1]
               totaltime+=delta_event
               if (len(self.bpmtable)>bpmindexT):
                  while (self.bpmtable[bpmindexT][0]<totaltime):        #not <= since we will calculate delta time before NOW and we need to do that with mpqn before NOW
#                     print "new mpqn: ",self.bpmtable[bpmindexT]," at: ",totaltime," Last mpqn: ",mpqnT
                     mpqnT=self.bpmtable[bpmindexT][1]  #60000000/mpqn    this is the bpm starting at that time
                     bpmindexT+=1
                     if (len(self.bpmtable)<=bpmindexT): break

#               beats_per_quarter_note = math.pow( 2, self.denominator )             # $ts_r is a 'time_signature' event (2**(@$ts_r->[3] - 2))
#               ticks_per_beat = delta_event/beats_per_quarter_note
#               ticks_per_second = (ticks_per_beat*1000000)/(mpqnT/beats_per_quarter_note)  #(Note that msecs is obtained by scaling by 1000 rather than 1000000).
#               ticks_per_second = ticks_per_second*50                                      #Calculate per 20us intervals.

               delta_time_in_microseconds=mpqnT*delta_event/self.ppqn      #for this event
#               if (delta_event):
#                  delta_time_in_microseconds=delta_event*1000000.0/mpqnT      #for this event
#               else:
#                  delta_time_in_microseconds=0
               totaltimeInMicrosecondsT+=delta_time_in_microseconds        #remember total delta time for all events (where are we in the track)
               deltaTimeNote_microsecsT=totaltimeInMicrosecondsT-totaltimeOld[channel]  #current total time - old totaltime for this channel
               totaltimeOld[channel]=totaltimeInMicrosecondsT                #remember time for next note
               #Calculate deltatime in milisecs and store in VBL bytes
               deltaTimeF=deltaTimeNote_microsecsT/1000.0*20.0                    #we need 20uS interval timing, we have time in uS, we need to divide by 20
               if (delta_event):
                 deltaTimeF+=0.5                                                     #Better rounding to intergers
               if (deltaTimeF):
                  if (deltaTimeF>maxdelta):
                     maxdelta=deltaTimeF
#                  flap=int((mpqnT*delta_event/self.ppqn/20.0)*100/deltaTimeF)
#                  if (flap!=flapold):
#                  print "deltatime org in 1000 x 20uS: ",deltaTimeF," According to new calc (with Time signature): ",ticks_per_second," org: ",mpqnT*delta_event/self.ppqn/1000.0
#                     print "deltatime org in uS: ",deltaTimeF," org: ",mpqnT*delta_event/self.ppqn/20.0," dif:: ",flap/100.0
#               print "deltatime org in 20 uS steps: ",deltaTimeF," maxdelta: ",maxdelta," int: ",int(deltaTimeF)
#                     flapold=flap

               if (deltaTimeNote_microsecsT%1000):     #lost uS
#                  print "Warning, losing time (uS):",deltaTimeNote_microsecsT%1000
                  lostuS+=deltaTimeNote_microsecsT%1000
                  if (deltaTimeNote_microsecsT%1000>=500):
#                     deltaTimeF+=1                      #half uS or more, add 1 ms for better rounding.
                     compensateduS+=1000
               deltaTimeAr=[]
   #            print "deltatime org in Ms=",deltaTimeMs
#               print deltaTimeF
               deltaTimeF=int(deltaTimeF)
#               print deltaTimeF
               if (deltaTimeF==0):
                  deltaTimeAr.append(0)      #in case of no time we want a 0 byte
               else:
                  while  (deltaTimeF):
                     deltaTimeAr.append(deltaTimeF%128)
                     if (len(deltaTimeAr)>1):   #Already 1 byte stored, so this and next VBR bytes have high bit set
                        deltaTimeAr[-1]+=128    #set MSB of last byte
                     deltaTimeF=deltaTimeF/128  #divide by 128
#               print deltaTimeAr
#               sys.exit()
#               print "track:",track," channel:",channel," note:",note[0],"Deltatime:",delta_event," in milisecs:",deltaTimeNote_microsecsT/1000," last mpqn: ",mpqnT,") totaltime in miliseconds=",totaltimeInMicrosecondsT/1000,"in tics=",totaltime," is ",totaltime/self.ppqn," quarter notes"
               if datalen[channel] < maxlen:    #Max 1000 notes per channel for now
                  if (bin):
                     dataout.append(note[0]) #note with msb=1 for on msb=0 for off
                     for byte in deltaTimeAr[::-1]:   #Loop backwards since msb is stored last
                        dataout.append(byte) # All time bytes in VBL (variable byte length (midi) in mS
                  else:
                     if start[channel]==0:
                        dataout+="  "+note[0] #note with msb=1 for on msb=0 for off
                        for byte in deltaTimeAr[::-1]:   #Loop backwards since msb is stored last
                           dataout+=",0x%02x" % byte # All time bytes in VBL (variable byte length (midi) in mS
                     else:
                        dataout+=","+note[0] #note with msb=1 for on msb=0 for off
                        for byte in deltaTimeAr[::-1]:   #Loop backwards since msb is stored last
                           dataout+=",0x%02x" % byte # All time bytes in VBL (variable byte length (midi) in mS
                     datalen[channel]+=1+len(deltaTimeAr)
                     start[channel]+=1+len(deltaTimeAr)
                     deltaTimeNote_microsecsT=0     #Start at 0 for next note
                     if start[channel]>15:
                        start[channel]=0
                        dataout+=",\n"
            self.closeTrack(track,channel,lostuS,compensateduS,dataout,datalen,outfilename,bin)

    def storeTrack(self,track,data):
      emptyChannels=1
      for channel in range(16):
         if channel in data:
            emptyChannels=0
            print "Track ",track," Channel ",channel," has data."
      if (not emptyChannels):
         self.trackdata[track]=data

    def combineChannels(self,combinetracks):
           if (not combinetracks==1): #combine some tracks
              for track in self.trackdata:
                 for channel in range(16):
                    if ((channel not in combinetracks) & (channel in self.trackdata[track]) ):
#                    if (channel not in combinetracks):
                       self.trackdata[track].pop(channel) #delete track if not in combine list
           keys = self.trackdata.keys()
           for key in keys:
              if (len(self.trackdata[key]) == 0):
                 self.trackdata.pop(key)

           #Combine the channels into 1
           combinedTrack=[]
           deltatime=0
           done=0
           doneflag=0
           TotalTimeNC=0
           TotalTimeC=0
           NrNotesNC=0
           NrNotesC=0
           for track in self.trackdata:
              for channel in self.trackdata[track]:
                 NrNotesNC+=len(self.trackdata[track][channel])
                 for note in self.trackdata[track][channel]:
                    TotalTimeNC+=note[1]
                 print "Separate Channel Nr: ",channel," Track: ",track," Total time: ",TotalTimeNC
                 TotalTimeNC=0
           print "Separate Channels Nr Notes: ",NrNotesNC
           while (not done):
              for track in self.trackdata:
                 for channel in self.trackdata[track]:
                    while (len(self.trackdata[track][channel])>0):
                       doneflag=1                                                            #loop at least one more time
                       if (self.trackdata[track][channel][0][1]==0):                              #DeltaTime for this note =0
                          combinedTrack.append([self.trackdata[track][channel][0][0],deltatime])  #we can write it to the combined track and the recorder time so far
                          TotalTimeC+=deltatime
                          NrNotesC+=1
                          if not NrNotesC%1000:
                             print "combining progress at note: ",NrNotesC
                          self.trackdata[track][channel].pop(0)                                   #delete note from the track(channel)
                          deltatime=0                                                        #and reset deltatime
                          if ((len(self.trackdata[track][channel])==0) & (not doneflag)):         #still notes left?
                             done=1                                                          #if not then done
                       else:                                                                 #Deltatime for this note != 0
                          break                                                              #Break out of while loop
                    else:
                       done=1
              if (doneflag):
                 done=0
                 doneflag=0
              if (not done):
                 for track in self.trackdata:
                    for channel in self.trackdata[track]:
                       if (len(self.trackdata[track][channel])>0):
                          self.trackdata[track][channel][0][1]-=1                                     #deltatime-- for this note
                 deltatime+=1                                               #deltatime++ for combined track
#           print "combined track: ",combinedTrack
           print "Combined channels Nr Notes: ",NrNotesC," Total time: ",TotalTimeC
           self.trackdata={0:{0:combinedTrack}}


    def run(self,infilename,outfilename,combinetracks):
        bin=self.bin
        reader=midireader(infilename)
        track=0
        for command,note,data1,delta_event,totalTime in reader:
            if (reader.chunk!=track):     #New track, clear all time data except bpm table
               if (reader.chunk>1):       #old track ended
                  self.storeTrack(track,data)                  #store all data from track and continue
               data={}
               totaltimeOld=[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]
               track=reader.chunk
               mpqnT=500000
               new_mpqnT=0
#               totalTimeInTics=0
#               totalTimeInTicsOld=0
               deltaTimeNote=0

#            delta_time_in_microseconds=delta_event                      #tics for this event
#            totalTimeInTics+=delta_event        #remember total delta time for all events (where are we in the track)
#            print "tticks: ",totalTimeInTics," ttotal: ",totaltime
            if (command==0xFF51):
               self.bpmtable.append([totalTime,data1]);    #will be totaldeltatime, mpqn
               print "bpmtable append at totaltime (in tics): ",totalTime," mpqn: ",data1;    #will be totaldeltatime, mpqn
               continue;
            if (command==0xFF58):
               self.numerator=note
               self.denominator=data1
               self.bar+=str(note)+"/"+str(data1)+", "
            channel,notedata,notelen = self.note2data(command,note,data1,bin)    #channel number, notedata (hex string) and number of bytes in the string are returned
            if not notedata: continue
            if (channel not in data):
               data[channel]=[];    #channel, empty data
#            if delta_time_in_microseconds > 0:
#            print "note:",notedata,"no. of quarter-notes (at 120bpm) for command:",hex(command)," =",no_of_quarts/16,deltaTimeNote/self.ppqn," remainder: ",deltaTimeNote%self.ppqn," (time in milisecs:",delta_time_in_microseconds/1000,") totaltime in miliseconds=",totaltimeInMicroseconds[channel]/1000,"in tics=",totaltime," is ",totaltime/self.ppqn," quarter notes"
#            print "track:",reader.chunk," channel:",channel," note:",notedata,"Deltatime:",deltaTimeNote," in milisecs:",delta_time_in_microseconds/1000," last mpqn: ",mpqn[channel],") totaltime in miliseconds=",totaltimeInMicroseconds[channel]/1000,"in tics=",totaltime," is ",totaltime/self.ppqn," quarter notes"
#            deltaTimeNote=totalTimeInTics-totaltimeOld[channel]  #current total time - old totaltime for this channel
            deltaTimeNote=totalTime-totaltimeOld[channel]  #current total time - old totaltime for this channel
            totaltimeOld[channel]=totalTime                     #remember time for next note
            data[channel].append([notedata,deltaTimeNote])                         #note with msb=1 for on msb=0 for off, for combinechannel this will be the note deta time in tics
        self.storeTrack(track,data)
        if (combinetracks):
           self.combineChannels(combinetracks)
        self.deltaTics2deltaMiliSecs(23000,outfilename)

class midi(Midi_Base):
#    def header0(self):
#        return '#ifndef _inc_midifile_\n#define _inc_midifile_\n\nPROGMEM prog_uint8_t midisong['
#    def header1(self):
#        return ']={\n'
#    def data_info(self):
#        return '  };\n// number of bytes (copy to array length field): '
#    def instrument_info(self):
#        return '\n// used instruments:'
#    def footer(self):
#        return '\n#endif /*_midifile_*/\n'
    def note2data(self, command,note,data1,bin):
        channel=command%16    #ls nibble = channel
        command=command/16    #ms nibble = command
#        if (command == 0x9):
        if ((command == 0x9) | (command == 0x8) ):						#note on  (0x9) or off (0x8) )message
#           return '0x0'+command+","+hex(note)+","+hex(data1)+','
#           print channel,hex(note),
           if (command == 0x9): note+=128 #note on if MSB=1, note off if MSB=0	# combine note nr. with on/off message (note=0..127, on=128 else off)
           if (bin):
              return channel,note,1
           else:
              return channel,hex(note),1
        elif (command == 0xC): #Program (patch) change cmd
#           print "Program (patch) change cmd, new program on channel",hex(byte%16)," instrument: ",program+1, self.instruments[program]
           self.instruments[channel]+=" "+data1
           return 0,'',0
        else:
           return 0,'',0
##        return 'D'+midi.date(row)+'\nT'+bank.amount(row)+'\nP'+bank.naam(row)+'\nM'+bank.omschrijving(row)+'\nN'+bank.tegenrek(row)+'\n^\n'     #Naam = P, omschrijving = M
##  0x04,etc



class midireader():
    "Iterator for looping over a midi  .mid file"
    def __init__(self,filename):
        self.reader = open(filename)
        self.reader.seek(14)  #Skip midifileheader
        self.chunk=0
        self.trackpos = 0 #how many bytes of track are read?
        self.totaltime = 0 #keep track of combined delta_time since start of track
        self.tracklen = 0 #Length of track chunk in bytes
        self.lastcommand = 0
        self.family=['Piano', 'Chromatic Percussion', 'Organ', 'Guitar', 'Bass', 'Strings', 'Ensemble', 'Brass', 'Reed', 'Pipe', 'Synth Lead', 'Synth Pad', 'Synth Effects', 'Ethnic', 'Percussive', 'Sound Effects']
        self.instruments=['Acoustic Grand Piano', 'Bright Acoustic Piano', 'Electric Grand Piano', 'Honky-tonk Piano', 'Electric Piano 1',
            'Electric Piano 2', 'Harpsichord', 'Clavinet', 'Celesta', 'Glockenspiel', 'Music Box', 'Vibraphone', 'Marimba', 'Xlophone',
            'Tubular Bells', 'Dulcimer', 'Drawbar Organ', 'Percussive Organ', 'Rock Organ', 'Church Organ', 'Reed Organ', 'Accordion',
            'Harmonica', 'Tango Accordion', 'Acoustic Guitar (nylon)', 'Acoustic Guitar (steel)', 'Electric Guitar (jazz)',
            'Electric Guitar (clean)', 'Electric Guitar (muted)', 'Overdriven Guitar', 'Distortion Guitar', 'Guitar harmonics',
            'Acoustic Bass', 'Electric Bass (finger)', 'Electric Bass (pick)', 'Fretless Bass', 'Slap Bass 1', 'Slap Bass 2',
            'Synth Bass 1', 'Synth Bass 2', 'Violin', 'Viola', 'Cello', 'Contrabass', 'Tremolo Strings', 'Pizzicato Strings',
            'Orchestral Harp', 'Timpani', 'String Ensemble 1', 'String Ensemble 2', 'Synth Strings 1', 'Synth Strings 2',
            'Choir Aahs', 'Voice Oohs', 'Synth Voice', 'Orchestra Hit', 'Trumpet', 'Trombone', 'Tuba', 'Muted Trumpet', 'French Horn',
            'Brass Section', 'Synth Brass 1', 'Synth Brass 2', 'Soprano Sax', 'Alto Sax', 'Tenor Sax', 'Baritone Sax', 'Oboe',
            'English Horn', 'Bassoon', 'Clarinet', 'Piccolo', 'Flute', 'Recorder', 'Pan Flute', 'Blown Bottle', 'Shakuhachi', 'Whistle',
            'Ocarina', 'Lead 1 (square)', 'Lead 2 (sawtooth)', 'Lead 3 (calliope)', 'Lead 4 (chiff)', 'Lead 5 (charang)',
            'Lead 6 (voice)', 'Lead 7 (fifths)', 'Lead 8 (bass + lead)', 'Pad 1 (new age)', 'Pad 2 (warm)', 'Pad 3 (polysynth)',
            'Pad 4 (choir)', 'Pad 5 (bowed)', 'Pad 6 (metallic)', 'Pad 7 (halo)', 'Pad 8 (sweep)', 'FX 1 (rain)', 'FX 2 (soundtrack)',
            'FX 3 (crystal)', 'FX 4 (atmosphere)', 'FX 5 (brightness)', 'FX 6 (goblins)', 'FX 7 (echoes)', 'FX 8 (sci-fi)', 'Sitar',
            'Banjo', 'Shamisen', 'Koto', 'Kalimba', 'Bag pipe', 'Fiddle', 'Shanai', 'Tinkle Bell', 'Agogo', 'Steel Drums', 'Woodblock',
            'Taiko Drum', 'Melodic Tom', 'Synth Drum', 'Reverse Cymbal', 'Guitar Fret Noise', 'Breath Noise', 'Seashore', 'Bird Tweet',
            'Telephone Ring', 'Helicopter', 'Applause', 'Gunshot']
    def __iter__(self):
        return self
    def memo(self):
        return self
    def next(self):
        if (self.trackpos >=self.tracklen):
           byte=self.reader.read(4)
           if (byte=="MTrk"):
              self.chunk+=1
              self.tracklen=0x01000000*ord(self.reader.read(1))
              self.tracklen+=0x010000*ord(self.reader.read(1))
              self.tracklen+=0x0100*ord(self.reader.read(1))
              self.tracklen+=ord(self.reader.read(1))
              print "\nTrack chunk ",self.chunk,". Length (bytes):",self.tracklen
           else:
              if byte:
                 print "Missing Track header...."
              else:
                 print "End of track and end of File detected."
                 raise StopIteration
                 return "",0,0,delta,self.totaltime
           self.trackpos = 0 #how many bytes of track are read?
           self.totaltime = 0
           self.lastcommand = 0
        byte = self.reader.read(1)
        self.trackpos+=1
        if byte:
           byte=ord(byte)
           delta=0
           while byte>128:    #delta time byte (always expected before a note or command. Last byte<128)
              if delta: delta=delta*128   #shift 7 bits to the right to fit next byte
              delta+=(byte-128)
              byte=ord(self.reader.read(1))
              self.trackpos+=1
           if delta: delta=delta*128   #shift 7 bits to the right to fit next byte
           delta+=byte        #last delta time byte
#           print "Deltatime=",delta
           self.totaltime+=delta
           byte=ord(self.reader.read(1))  #command or note
           self.trackpos+=1
           while byte<128:
              byte=self.lastcommand
#              print "No command, assume last command for this byte (running mode):",hex(byte)
              self.reader.seek(-1,1)   #filepointer -1 byte (to current pos (=-1))
              self.trackpos-=1         #subtrackt it from trackpos too
           if byte>127:
              self.lastcommand=byte
              if byte==0xFF:
                 byte = ord(self.reader.read(1))  #command of meta event
                 self.trackpos+=1
#                 print "Meta event detected: 0xFF",hex(byte),"=(",byte,")"
                 if (byte==0):
                    self.reader.read(1) # = 02
                    print "track sequence number: ",ord(self.reader.read(1))*256+ord(self.reader.read(1))
                    self.trackpos+=3
                 if ((byte>0) & (byte<8)):
                    txtlen = ord(self.reader.read(1))  #length of text
                    txt=self.reader.read(txtlen)  #the text
                    self.trackpos+=txtlen+1
                    if (txtlen > 3):
                       print "Meta text event: "+txt
                 elif (byte==0x2f):
                    print "End of track byte detected"
                    self.reader.read(1)
                    self.trackpos+=1
                 elif (byte==0x51):
                    self.reader.read(1) #will be 3 for 3 bytes of data
                    mpqn=ord(self.reader.read(1))*65536+ord(self.reader.read(1))*256+ord(self.reader.read(1))
#                    print "Set tempo at bytepos:",self.trackpos,", timepos:",self.totaltime,", Microseconds Per Quarter Note",mpqn," (=",60000000/mpqn,")"
                    self.trackpos+=4
                    return 0xFF51,0,mpqn,delta,self.totaltime
                 elif (byte==0x54):
                    self.reader.read(1) #will be 5 for 5 bytes of data
                    hour=ord(self.reader.read(1))
                    framerate=hour/64
                    if framerate>=2: framerate=30
                    else: framerate+=24
                    hour=hour%64
                    print "SMPTE offset, framerate:",framerate,", time:",hour,":",ord(self.reader.read(1)),":",ord(self.reader.read(1))," framenr:",ord(self.reader.read(1))," subframenr (always 100/frame):",ord(self.reader.read(1))
                    self.trackpos+=6
                 elif (byte==0x58):
                    self.reader.read(1)
                    numerator=ord(self.reader.read(1))
                    denominator=ord(self.reader.read(1))
                    print "Time Signature. numerator:",numerator," denominator:",denominator," nr. ticks in metronome click:",ord(self.reader.read(1)),"nr. of 32nd notes to the quarter note",ord(self.reader.read(1))
                    self.trackpos+=5
                    return 0xFF58,numerator,denominator,delta,self.totaltime
                 elif (byte==0x59):
                    self.reader.read(1)
                    print "Key Signature. sharp/flats:",ord(self.reader.read(1))," major/minor:",ord(self.reader.read(1))
                    self.trackpos+=3
                 elif (byte==0x7F):
                    datalen = ord(self.reader.read(1))  #length of data
                    data=self.reader.read(datalen)  #the data
                    self.trackpos+=datalen+1
                    print "Sequencer specific information (nr. of bytes): ",datalen
                 else:
                    datalen = ord(self.reader.read(1))  #length of data
                    data=self.reader.read(datalen)  #the data
                    self.trackpos+=datalen+1
                    print "Unknown Meta event, some data not decoded... code:",hex(byte)," len:",hex(datalen)," data:",data
                 return 0xFF,0,byte,delta,self.totaltime
              elif ((byte>=0x80) & (byte<0x90)): #note off cmd
                 note = ord(self.reader.read(1))
                 velocity = ord(self.reader.read(1))
                 self.trackpos+=2
#                 return byte,note,velocity,self.totaltime
                 return byte,note,0,delta,self.totaltime
              elif ((byte>=0x90) & (byte<0xA0)): #note on cmd
                 note = ord(self.reader.read(1))
                 velocity = ord(self.reader.read(1))
                 if ((velocity == 0)): byte -=0x10     #note on (0x9n) with volume 0 means note off (0x8n)
                 self.trackpos+=2
#                 if ((self.trackpos<=100) | (self.chunk==0)):
#                    print "channel nr:",byte%16
                 return byte,note,0,delta,self.totaltime
              elif ((byte>=0xA0) & (byte<0xB0)): #key after touch cmd
                 print "key after touch cmd"
                 note = ord(self.reader.read(1))
                 data1 = ord(self.reader.read(1))
                 self.trackpos+=2
                 return byte,note,data1,delta,self.totaltime
              elif ((byte>=0xB0) & (byte<0xC0)): #control change cmd
                 controller = ord(self.reader.read(1))
                 value = ord(self.reader.read(1))
                 self.trackpos+=2
#                 print "control change cmd. controller: ",hex(controller)," value:",hex(value)
                 return byte,controller,value,delta,self.totaltime
              elif ((byte>=0xC0) & (byte<0xD0)): #Program (patch) change cmd
                 program = ord(self.reader.read(1))
                 self.trackpos+=1
                 print "Program (patch) change cmd, new program on channel",hex(byte%16),"family=",self.family[program/8]," instrument=",program+1, self.instruments[program]
                 return byte,program,str(program+1)+": family="+self.family[program/8]+" instrument="+self.instruments[program],delta,self.totaltime
              elif ((byte>=0xD0) & (byte<0xE0)): #Channel after touch cmd
                 print "Channel after touch cmd"
                 channel = ord(self.reader.read(1))
                 self.trackpos+=1
                 return byte,channel,0,delta,self.totaltime
              elif ((byte>=0xE0) & (byte<0xF0)): #Pitch wheel change cmd
                 bottom = ord(self.reader.read(1))
                 top = ord(self.reader.read(1))
                 self.trackpos+=2
#                 print "Pitch wheel change cmd. Bottom="+hex(bottom)+" Top="+hex(top)
                 return byte,bottom,top,delta,self.totaltime
              elif (byte==0xF0):
                 print "System Message: System Exclusive "+hex(byte)
              elif (byte==0xF1):
                 print "Time Code:"+hex(ord(self.reader.read(1)))
                 self.trackpos+=1
              elif (byte==0xF2):
                 lsb=hex(ord(self.reader.read(1)))
                 print "Song position pointer:"+hex(ord(self.reader.read(1)))+lsb
                 self.trackpos+=2
              elif (byte==0xF3):
                 print "Song Select:"+hex(ord(self.reader.read(1)))
                 self.trackpos+=1
              elif (byte==0xF4):
                 print "System Message: undefined:"+hex(byte)
              elif (byte==0xF5):
                 print "System Message: undefined:"+hex(byte)
              elif (byte==0xF6):
                 print "System Message: Tune Request"
              elif (byte==0xF7):
                 print "System Message: EOX (End of System Exclusive)"
              elif (byte==0xF8):
                 print "Realtime Message: Clock"
              elif (byte==0xF9):
                 print "Realtime Message: Undefined"
              elif (byte==0xFA):
                 print "Realtime Message: Start"
              elif (byte==0xFB):
                 print "Realtime Message: Continue"
              elif (byte==0xFC):
                 print "Realtime Message: Stop"
              elif (byte==0xFD):
                 print "Realtime Message: Undefined"
              elif (byte==0xFE):
                 print "Realtime Message: Active Sensing (we are still alive!)"
              elif (byte==0xFF):
                 print "Realtime Message: System Reset"
              else:
                 print "Unknown command ************:"+hex(byte)
              return byte,0,0,delta,self.totaltime
           else:
              print "we should not come to here!"
        else:
           print "End of file detected before end of track (according to specified length from header)."
           raise StopIteration
           return "",0,0,delta,self.totaltime

def no_combine():
   if (sys.argv[1]=="-b"):
       return 1,sys.argv[2],0
   elif (sys.argv[1]=="-h"):
       return 0,sys.argv[2],0
   else:
      print '\nUnknown option, please provide -b or -h as first argument.\n'
      return 0,0,0
def combine_all():
   if (sys.argv[2]=="-c"):
      print 'combining all available channels.\n'
      if (sys.argv[1]=="-b"):
          return 1,sys.argv[3],1
      elif (sys.argv[1]=="-h"):
          return 0,sys.argv[3],1
      else:
          print '\nUnknown option, please provide -b or -h as first argument.\n'
          return 0,0,0
   else:
      print '\nUnknown option, please provide a valid option (first argument -b or -h, second argument (optional) -c)).\n'
      return 0,0,0
def combine_list():
   if (sys.argv[1]=="-b"):
       binary=1
   elif (sys.argv[1]=="-h"):
       binary=0
   else:
      print '\nUnknown option, please provide -b or -h as first argument.\n'
      return 0,0,0
   if (sys.argv[2]=="-c"):
      list=sys.argv[3].split(',')
      if (len(list)<2):
         print '\nProvide at least 2 channel numbers to combine.\n'
         sys.exit()
      else:
         for key,number in enumerate(list):
            if number.isdigit():
               list[key]=int(number)
            else:
               print '\nChannel numbers should be valid numbers.\n'
               sys.exit()
            if ((int(number)<0) | (int(number)>15)):
               print '\nChannel numbers should be between 0 and 15.\n'
               sys.exit()
      print 'combining channels: ',list,'\n'
      return binary,sys.argv[4],list
   else:
      print '\nUnknown option, please provide a valid option (first argument -b or -h, second argument (optional) -c)).\n'
      return 0,0,0

if __name__ == '__main__':
   import sys
   help = '\nUsage: midi2mmi "filename" to convert a file\n\nOptions:\n -[b|h] -c [n,n,n,n...]    -b or -h = binary or include file output. -c = combine channels (all channels if not specified by komma separated list.)\n\n'
   switch = {3: no_combine,
    4: combine_all,
    5: combine_list}
   if len(sys.argv) in switch:
      bin,inputfilename,combine=switch[len(sys.argv)]()
   else:     #default
#    if len(sys.argv) > 4:
      print help
      sys.exit()
   if (inputfilename==0):
      print help
      sys.exit()
#   print inputfilename,combine
   if inputfilename[-4:] != '.mid':
      outputfilename=inputfilename
      inputfilename += '.mid'
   else:
      outputfilename=inputfilename[:-4]
   #print '\nfilename:     ' + inputfilename + '\n'
  # print '\ncombine:      ' + `combine` + '\n'
   #print '\nbin:          ' + `bin` + '\n'
  # sys.exit()
   print '\nLezen:     ' + inputfilename + '\n\n'
   instance = midi(inputfilename,bin)
   instance.run(inputfilename,outputfilename,combine)
   print 'Schrijven: ' + outputfilename + ' gelukt. '
